require 'devrandom'

class Bonkle
	class RandomDice
		# Get a set of dice
		def initialize
			@dr = DevRandom.new
			reset
			self
		end
		# Reinitialize to a full set of dice
		def reset
			@dice = [
				%w(A A E E G N),
				%w(A B B J O O),
				%w(A C H O P S),
				%w(A F F K P S),
				%w(A O O T T W),
				%w(C I M O T U),
				%w(D E I L R X),
				%w(D E L R V Y),
				%w(D I S T T Y),
				%w(E I O S S T),
				%w(E E I N S U),
				%w(E E G H N W),
				%w(E H R T V W),
				%w(E L R T T Y),
				%w(H I M N Qu U),
				%w(H L N N R Z),
			]
			true
		end	
		# Take random dice from the set.  With no arguments or passed 1,
		# return the single string from one die, else an Array of strings
		# of the number of dice asked for.
		#
		# This raises an EOFError exception if too many dice are asked for.
		def grab(count = 1)
			if @dice.size == 0
				raise EOFError, 'selection of dice empty'
			end
			case count
			when 2 .. @dice.size
				ret = []
				count.downto(1) {
					pos = @dr.read(4) % @dice.size
					die = @dice.delete_at(pos)
					ret.push die[@dr.read(4) % 6]
				}
				return ret
			when 1
				pos = @dr.read(4) % @dice.size
				die = @dice.delete_at(pos)
				return die[@dr.read(4) % 6]
			else
				raise ArgumentError, 'bad die count'
			end
		end
		# Are there dice left?
		def empty?
			@dice.empty?
		end
		# Are there all the dice left?
		def full?
			@dice.size == 16
		end
	end
	class Board
		# Create a new "board" of random dice
		def initialize
			@dice = Bonkle::RandomDice.new
			reset
			self
		end
		# Reinitialize the board to a new random one
		def reset
			@board = Internal.new
			@dice.reset if !@dice.full?
			4.times {
				@board.push @dice.grab(4)
			}
			freeze_board
			true
		end
		class Internal < Array
			def find_letter(letter)
				if $DEBUG
					printf("find_letter(\"%s\")\n", letter)
				end
				locations = []
				self.each_with_index {|row, rowi|
					row.each_with_index {|die, diei|
						if letter == die
							locations.push [rowi, diei]
						end
					}
				}
				locations
			end
			# Take an array of letters and find the next one.
			def find_word(word, prev_coords)
				if $DEBUG
					printf("find_word(\"%s\", [%d, %d])\n", word, *prev_coords)
				end
				return true if word.empty?
				self.find_letter(word[0]).each {|coords|
					if (coords[0] - prev_coords[0]).between?(-1, 1) &&
					   (coords[1] - prev_coords[1]).between?(-1, 1)
						nb = self.dup
						nb[*coords].concat '_'
						return true if nb.find_word word[1..-1], coords
					end
				}
				false
			end
			def dup
				b = super
				b.each_index {|rowi|
					b[rowi] = b[rowi].dup
					b[rowi].each_index {|diei|
						b[rowi][diei] = b[rowi][diei].dup
					}
				}
				b
			end
			def [](*x)
				if x.length == 2
					super(x[0])[x[1]]
				else
					super(*x)
				end
			end
		end
		def [](*x)
				@board[*x]
		end
		attr_reader :board
		def has_word?(word)
			if word =~ /q[^u]/i || word !~ /^[a-z]{3,17}$/i
				return false
			end
			myword = []
			word.upcase!
			word.gsub! /QU/, 'Qu'
			word.gsub(/[A-Z]u?/) {|x| myword.push x}
			@board.find_letter(myword[0]).each {|coords|
				nb = @board.dup
				nb[*coords].concat '_'
				return true if nb.find_word myword[1..-1], coords
			}
			false
		end
		def to_s
			border = "+---+---+---+---+\n"
			rowfmt = "| %-2s" * 4 + "|\n" + border
			string = border.dup
			@board.each {|row|
				string.concat format(rowfmt, *row)
			}
			string
		end
		def freeze
			freeze_board
			super
		end
		private
		def freeze_board
			@board.each {|row|
				row.each {|die|
					die.freeze
				}
				row.freeze
			}
			@board.freeze
		end
	end
end
